@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2004, 2005 Marcus Glocker <marcus@nazgul.ch>
+ * Copyright (c) 2004 - 2022 Marcus Glocker <marcus@nazgul.ch>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -54,7 +54,7 @@ strb64d(char *dst, const char *src, const int dsize)
 	dst[0] = '\0';
 
 	while (1) {
-		for (i = 0, j = j; i < 4; i++, j++) {
+		for (i = 0; i < 4; i++, j++) {
 			if (src[j] == '\0')
 				goto quit;
 			ch = src[j];
