@@ -1,5 +1,7 @@
+/*	$nostromo: extern.h,v 1.65 2016/04/12 19:02:06 hacki Exp $ */
+
 /*
- * Copyright (c) 2004 - 2011 Marcus Glocker <marcus@nazgul.ch>
+ * Copyright (c) 2004 - 2016 Marcus Glocker <marcus@nazgul.ch>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -228,3 +230,14 @@ struct listener {
 #define IP6SSL	16			/* Set for IPv6 SSL socket */
 };
 extern struct listener l[];
+
+/*
+ * basic authentication cache
+ */
+struct ba_cache {
+	time_t	timestamp;		/* Timestamp when entry was added */
+	char	username[1024];		/* Username */
+	char	password[1024];		/* Password in clear text! */
+#define BAC_ENTRIES 128
+};
+extern struct ba_cache **bac;
