@@ -1,5 +1,7 @@
+/*	$nostromo: http.c,v 1.172 2022/12/15 19:18:41 hacki Exp $ */
+
 /*
- * Copyright (c) 2004 - 2011 Marcus Glocker <marcus@nazgul.ch>
+ * Copyright (c) 2004 - 2016 Marcus Glocker <marcus@nazgul.ch>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -34,8 +36,8 @@
 #include "../libmy/str.h"
 #include "../libmy/file.h"
 #ifdef __linux__
-#include "../libbsd/strlcpy.h"
-#include "../libbsd/strlcat.h"
+#include "../libmy/strlcpy.h"
+#include "../libmy/strlcat.h"
 #endif
 #include "config.h"
 #include "proto.h"
@@ -145,7 +147,8 @@ http_verify(char *header, const int header_size, const char *cip, const int sfd,
 	}
 
 	/* check for valid uri */
-	if (strstr(header, "/../") != NULL) {
+	if (strstr(header, "/../") != NULL ||
+	    (config.homedirs[0] != '\0' && strstr(header, "/~../") != NULL)) {
 		h = http_head(http_s_400, line, cip, 0);
 		b = http_body(http_s_400, "", h, 0);
 		c[sfd].pfdo++;
@@ -448,6 +451,9 @@ http_proc(const char *header, char *body, const int hr, const int blen,
 		}
 		x_fork = 1;
 
+		/* nhttpd child processes follow default signal handling */
+		sys_sighandler(0);
+
 		/* close all copied fds from parent */
 		sys_close_except(sfd);
 
@@ -455,8 +461,16 @@ http_proc(const char *header, char *body, const int hr, const int blen,
 		sys_resetnonblock(sfd);
 
 		/* create pipes to communicate with cgi */
-		pipe(fds1);
-		pipe(fds2);
+		if (pipe(fds1) == -1) {
+			syslog(LOG_ERR, "can't fork cgi: pipe fds1: %s",
+			    strerror(errno));
+			exit(1);
+		}
+		if (pipe(fds2) == -1) {
+			syslog(LOG_ERR, "can't fork cgi: pipe fds2: %s",
+			    strerror(errno));
+			exit(1);
+		}
 
 		/* fork child for cgi */
 		if ((cpid = fork()) == -1) {
@@ -471,14 +485,18 @@ http_proc(const char *header, char *body, const int hr, const int blen,
 			close(fds1[1]);
 			close(fds2[0]);
 
-			chdir(rh->rq_filep);
+			if (chdir(rh->rq_filep) == -1) {
+				syslog(LOG_ERR, "can't fork cgi: chdir: %s",
+				    strerror(errno));
+				exit(1);
+			}
 			dup2(fds1[0], STDIN_FILENO);
 			dup2(fds2[1], STDOUT_FILENO);
 
 			/* build cgi environment array */
 			i = 0;
 			cgiarg[0] = rh->rq_files;
-			cgiarg[1] = '\0';
+			cgiarg[1] = NULL;
 			cgienv[i++] = sys_benv("DOCUMENT_ROOT=%s",
 			    rh->rq_docrootv);
 			cgienv[i++] = sys_benv("GATEWAY_INTERFACE=%s",
@@ -559,7 +577,7 @@ http_proc(const char *header, char *body, const int hr, const int blen,
 			else
 				cgienv[i++] = sys_benv("SERVER_PORT=%s",
 				    config.serverportc);
-			cgienv[i++] = '\0';
+			cgienv[i++] = NULL;
 
 			execve(rh->rq_filef, cgiarg, cgienv);
 			exit(0);
@@ -761,6 +779,9 @@ http_proc(const char *header, char *body, const int hr, const int blen,
 		}
 		x_fork = 1;
 
+		/* nhttpd child processes follow default signal handling */
+		sys_sighandler(0);
+
 		/* close all copied fds from parent */
 		sys_close_except(sfd);
 
@@ -1054,15 +1075,15 @@ http_cgi_header(char *header_cgi, char *status, const int dsize1,
  * http_header_comp()
  * 	check if received headers arrived complete
  * Return:
- * 	0 = headers not complete, 1 = headers complete
+ * 	0 = headers not complete, <number of headers> = headers complete
  */
 int
 http_header_comp(char *header, const int len)
 {
-	int	r;
-	char	*p, *end;
+	int	i, headers;
+	char	*p;
 
-	r = 0;
+	headers = 0;
 
 	/* check header for minimum size */
 	if (len < 4)
@@ -1078,11 +1099,18 @@ http_header_comp(char *header, const int len)
 	}
 
 	/* any header */
-	end = header + (len - 4);
-	if (!strcmp(end, "\r\n\r\n"))
-		r = 1;
+	for (i = 0; i < len; i++) {
+		if (header[i] == '\r') {
+			if ((len - i) < 4)
+				break;
+			if (!strncmp(&header[i], "\r\n\r\n", 4)) {
+				headers++;
+				i += 3;
+			}
+		}
+	}
 
-	return (r);
+	return (headers);
 }
 
 /*
@@ -1575,7 +1603,7 @@ http_header(const char *header_data, const char *force_status,
 	    strcmp(http_urls, h->rq_fv_hos) != 0) {
 		/* search for virutal host in configuration */
 		if (fparse(h->rq_docrootv, h->rq_fv_hos, config.file,
-		    sizeof(h->rq_docrootv)) > 0) {
+		    sizeof(h->rq_docrootv), 0) > 0) {
 			/* set own access_log for virutal host */
 			snprintf(logn, sizeof(logn), "%s_%s", config.logaccess,
 			    h->rq_fv_hos);
@@ -1588,7 +1616,7 @@ http_header(const char *header_data, const char *force_status,
 		strlcpy(alias, h->rq_uri, sizeof(alias));
 
 	/* check if uri contains an alias */
-	if (fparse(docroot, alias, config.file, sizeof(docroot)) != -1) {
+	if (fparse(docroot, alias, config.file, sizeof(docroot), 0) != -1) {
 		/* yes cut it out */
 		if (strcuti(file_path, h->rq_uri, strlen(alias),
 		    strlen(h->rq_uri), sizeof(file_path)) == -1)
@@ -1801,7 +1829,7 @@ http_header(const char *header_data, const char *force_status,
 #endif
 				/* get access realm */
 				if (fparse(h->rp_fv_auw, "realm", acc_file,
-				    sizeof(h->rp_fv_auw)) == -1) {
+				    sizeof(h->rp_fv_auw), 0) == -1) {
 					/* realm could not be parsed */
 					strlcpy(h->rp_fv_auw, "unknown realm",
 					    sizeof(h->rp_fv_auw));
