@@ -1,5 +1,7 @@
+/*	$nostromo: main.c,v 1.228 2022/12/15 19:18:41 hacki Exp $ */
+
 /*
- * Copyright (c) 2004 - 2011 Marcus Glocker <marcus@nazgul.ch>
+ * Copyright (c) 2004 - 2016 Marcus Glocker <marcus@nazgul.ch>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -45,8 +47,8 @@
 #include "../libmy/str.h"
 #include "../libmy/file.h"
 #ifdef __linux__
-#include "../libbsd/strlcpy.h"
-#include "../libbsd/strlcat.h"
+#include "../libmy/strlcpy.h"
+#include "../libmy/strlcat.h"
 #endif
 #include "config.h"
 #include "proto.h"
@@ -72,6 +74,7 @@ int			fdbsdauth[2];
 #endif
 char			*mimes;
 char			logn[1024];
+struct ba_cache		**bac;
 
 const char *month[12] = {
 	"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct",
@@ -114,7 +117,7 @@ const char *http_fn_auw		= "WWW-Authenticate:";
 const char *http_fn_ran		= "Range:";
 const char *http_fn_rac		= "Content-Range:";
 const char *http_fn_aen		= "Accept-Encoding:";
-const char *http_fv_srv		= "nostromo 1.9.5";
+const char *http_fv_srv		= "nostromo 2.1";
 const char *http_fv_pr0	=	 "HTTP/1.0";
 const char *http_fv_pr1	=	 "HTTP/1.1";
 const char *http_fv_cgi		= "CGI/1.1";
@@ -181,26 +184,20 @@ sig_handler(const int sig)
 	switch (sig) {
 	case SIGCHLD: {
 		int	status;
+		int	save_errno = errno;
 
 		while (waitpid(-1, &status, WNOHANG) > 0);
 
+		errno = save_errno;
+
 		break;
 	}
 	case SIGINT:
 	case SIGTERM:
 		quit = 1;
 		break;
-	case SIGPIPE:
-		/* ignore */
-		break;
 	case SIGHUP:
-		/* ignore */
-		break;
-	case SIGQUIT:
-		/* ignore */
-		break;
-	case SIGALRM:
-		/* ignore */
+		load_config(1);
 		break;
 	default:
 		syslog(LOG_INFO, "unhandled signal %i, ignored", sig);
@@ -415,6 +412,192 @@ sdlisten(int sd, int sdtype)
 	return (1);
 }
 
+/*
+ * load_config()
+ *	load or reload configuration file
+ * Return:
+ *	0 = success, -1 = failed
+ */
+int
+load_config(int reload)
+{
+	int r = 0;
+	char tmp[1024];
+
+	/* optional: 'logaccess' - reloadable */
+	if (fparse(config.logaccess, "logaccess", config.file,
+	    sizeof(config.logaccess), 0) == -1) {
+		config.logaccess_flag = 0;
+	} else {
+		config.logaccess_flag = 1;
+		strlcpy(logn, config.logaccess, sizeof(logn));
+	}
+
+	/* optional: 'htaccess' - reloadable */
+	if (fparse(config.htaccess, "htaccess", config.file,
+	    sizeof(config.htaccess), 0) == -1)
+		config.htaccess[0] = '\0';
+
+	/* optional: 'custom_401' - reloadable */
+	if (fparse(config.c401, "custom_401", config.file,
+	    sizeof(config.user), 0) == -1)
+		config.c401[0] = '\0';
+
+	/* optional: 'custom_403' - reloadable */
+	if (fparse(config.c403, "custom_403", config.file,
+	    sizeof(config.user), 0) == -1)
+		config.c403[0] = '\0';
+
+	/* optional: 'custom_403' - reloadable */
+	if (fparse(config.c404, "custom_404", config.file,
+	    sizeof(config.user), 0) == -1)
+		config.c404[0] = '\0';
+
+	/* optional: 'homedirs' - reloadable */
+	if (fparse(config.homedirs, "homedirs", config.file,
+	    sizeof(config.homedirs), 0) == -1)
+		config.homedirs[0] = '\0';
+
+	/* optional: 'homedirs_public' - reloadable */	
+	if (fparse(config.homedirs_public, "homedirs_public", config.file,
+	    sizeof(config.homedirs_public), 0) == -1) {
+		config.homedirs_public[0] = '\0';
+	} else {
+		if (config.homedirs_public[0] != '/') {
+			strlcpy(tmp, config.homedirs_public, sizeof(tmp));
+			snprintf(config.homedirs_public,
+			    sizeof(config.homedirs_public), "/%s", tmp);
+		}
+	}
+
+	/* if we are in reload mode we leave here because the following
+	 * parameters are not reloadable */
+	if (reload) {
+		/* clear Basic Authentication Cache as well */
+		sys_bac_init();
+		syslog(LOG_INFO, "configuration has been reloaded");
+		return (r);
+	}
+
+	/* mandatory: 'servername' */
+	if (fparse(config.servername, "servername", config.file,
+	    sizeof(config.servername), 0) == -1) {
+		fprintf(stderr, "Missing <servername> in config file.\n");
+		r = -1;
+	} else {
+		/* get servers port */
+		if (strcuts(config.serverportc, config.servername, ':', '\0',
+		    sizeof(config.serverportc)) == -1) {
+			/* default port */
+			config.serverport = PRT;
+			snprintf(config.serverportc,
+			    sizeof(config.serverportc), "%d", PRT);
+		} else {
+			/* custom port */
+			config.serverport = atoi(config.serverportc);
+		}
+
+		/* remove port from servername */
+		strcuts(config.servername, config.servername, '\0', ':',
+		    sizeof(config.servername));
+	}
+
+	/* mandatory: 'serverlisten' */
+	if (fparse(config.serverlisten, "serverlisten", config.file,
+	    sizeof(config.serverlisten), 0) == -1) {
+		fprintf(stderr, "Missing <serverlisten> in config file.\n");
+		r = -1;
+	}
+
+	/* mandatory: 'serveradmin' */
+	if (fparse(config.serveradmin, "serveradmin", config.file,
+	    sizeof(config.serveradmin), 0) == -1) {
+		fprintf(stderr, "Missing <serveradmin> in config file.\n");
+		r = -1;
+	}
+
+	/* mandatory: 'serverroot' */
+	if (fparse(config.serverroot, "serverroot", config.file,
+	    sizeof(config.serverroot), 0) == -1) {
+		fprintf(stderr, "Missing <serverroot> in config file.\n");
+		r = -1;
+	}
+
+	/* mandatory: 'servermimes' */
+	if (fparse(config.servermimes, "servermimes", config.file,
+	    sizeof(config.servermimes), 0) == -1) {
+		fprintf(stderr, "Missing <servermimes> in config file.\n");
+		r = -1;
+	}
+
+	/* mandatory: 'docroot' */
+	if (fparse(config.docroot, "docroot", config.file,
+	    sizeof(config.docroot), 0) == -1) {
+		fprintf(stderr, "Missing <docroot> in config file.\n");
+		r = -1;
+	}
+
+	/* mandatory: 'docindex' */
+	if (fparse(config.docindex, "docindex", config.file,
+	    sizeof(config.docindex), 0) == -1) {
+		fprintf(stderr, "Missing <docindex> in config file.\n");
+		r = -1;
+	}
+
+	/* optional: 'logpid' */
+	if (fparse(config.logpid, "logpid", config.file,
+	    sizeof(config.logpid), 0) == -1) {
+		config.logpid_flag = 0;
+	} else {
+		config.logpid_flag = 1;
+	}
+
+	/* optional: 'user' */
+	if (fparse(config.user, "user", config.file,
+	    sizeof(config.user), 0) == -1)
+		config.user[0] = '\0';
+
+	/* optional: 'htpasswd' */
+	if (fparse(config.htpasswd, "htpasswd", config.file,
+	    sizeof(config.htpasswd), 0) == -1)
+		config.htpasswd[0] = '\0';
+#ifdef __OpenBSD__
+	else {
+		if (strcmp(config.htpasswd, "+bsdauth") == 0) {
+			config.bsdauth = 1;
+		} else if (strcmp(config.htpasswd, "+bsdauthnossl") == 0) {
+			config.bsdauth = 2;
+		} else {
+			config.bsdauth = 0;
+		}
+	}
+#endif
+	/* optional: 'sslport' */
+	if (fparse(config.sslportc, "sslport", config.file,
+	    sizeof(config.sslportc), 0) == -1) {
+		config.ssl = 0;
+	} else {
+		config.ssl = 1;
+		config.sslport = atoi(config.sslportc);
+
+		if (fparse(config.sslcert, "sslcert", config.file,
+		    sizeof(config.sslcert), 0)  == -1) {
+			fprintf(stderr,
+			    "Missing <sslcert> in config file.\n");
+			r = -1;
+		}
+
+		if (fparse(config.sslcertkey, "sslcertkey", config.file,
+		    sizeof(config.sslcertkey), 0) == -1) {
+			fprintf(stderr,
+			    "Missing <sslcertkey> in config file.\n");
+			r = -1;
+		}
+	}
+
+	return (r);
+}
+
 /*
  * main()
  *	nostromo webserver
@@ -490,149 +673,33 @@ main(int argc, char *argv[])
 			/* NOTREACHED */
 		}
 	}
+
+	/*
+	 * process configuration file.
+	 */
+
 	/* if no custom configuration file is given take the default */
 	if (!opt_c)
 		strlcpy(config.file, configfile, sizeof(config.file));
+
 	/* check if configuration file exists */
-	if ((fd = open(config.file, O_RDONLY)) == -1)
+	fd = open(config.file, O_RDONLY);
+	if (fd == -1)
 		err(1, "%s", config.file);
 	else
 		close(fd);
 
-	/*
-	 * parse configuration file
-	 */
-
-	/* mandatory configuration parameters */
-	if (fparse(config.servername, "servername", config.file,
-	    sizeof(config.servername)) == -1) {
-		fprintf(stderr, "Missing <servername> in config file.\n");
-		r = -1;
-	} else {
-		/* get servers port */
-		if (strcuts(config.serverportc, config.servername, ':', '\0',
-		    sizeof(config.serverportc)) == -1) {
-			/* default port */
-			config.serverport = PRT;
-			snprintf(config.serverportc, sizeof(config.serverportc),
-			    "%d", PRT);
-		} else
-			/* custom port */
-			config.serverport = atoi(config.serverportc);
-
-		/* remove port from servername */
-		strcuts(config.servername, config.servername, '\0', ':',
-		    sizeof(config.servername));
-	}
-	if (fparse(config.serverlisten, "serverlisten", config.file,
-	    sizeof(config.serverlisten)) == -1) {
-		fprintf(stderr, "Missing <serverlisten> in config file.\n");
-		r = -1;
-	}
-	if (fparse(config.serveradmin, "serveradmin", config.file,
-	    sizeof(config.serveradmin)) == -1) {
-		fprintf(stderr, "Missing <serveradmin> in config file.\n");
-		r = -1;
-	}
-	if (fparse(config.serverroot, "serverroot", config.file,
-	    sizeof(config.serverroot)) == -1) {
-		fprintf(stderr, "Missing <serverroot> in config file.\n");
-		r = -1;
-	}
-	if (fparse(config.servermimes, "servermimes", config.file,
-	    sizeof(config.servermimes)) == -1) {
-		fprintf(stderr, "Missing <servermimes> in config file.\n");
-		r = -1;
-	}
-	if (fparse(config.docroot, "docroot", config.file,
-	    sizeof(config.docroot)) == -1) {
-		fprintf(stderr, "Missing <docroot> in config file.\n");
-		r = -1;
-	}
-	if (fparse(config.docindex, "docindex", config.file,
-	    sizeof(config.docindex)) == -1) {
-		fprintf(stderr, "Missing <docindex> in config file.\n");
-		r = -1;
-	}
-
-	/* optional configuration parameters */
-	if (fparse(config.logpid, "logpid", config.file,
-	    sizeof(config.logpid)) == -1) {
-		config.logpid_flag = 0;
-	} else {
-		config.logpid_flag = 1;
-	}
-	if (fparse(config.logaccess, "logaccess", config.file,
-	    sizeof(config.logaccess)) == -1) {
-		config.logaccess_flag = 0;
-	} else {
-		config.logaccess_flag = 1;
-		strlcpy(logn, config.logaccess, sizeof(logn));
-	}
-	if (fparse(config.user, "user", config.file, sizeof(config.user))
-	    == -1)
-		config.user[0] = '\0';
-	if (fparse(config.htaccess, "htaccess", config.file,
-	    sizeof(config.htaccess)) == -1)
-		config.htaccess[0] = '\0';
-	if (fparse(config.htpasswd, "htpasswd", config.file,
-	    sizeof(config.htpasswd)) == -1)
-		config.htpasswd[0] = '\0';
-#ifdef __OpenBSD__
-	else {
-		if (strcmp(config.htpasswd, "+bsdauth") == 0)
-			config.bsdauth = 1;
-		else if (strcmp(config.htpasswd, "+bsdauthnossl") == 0)
-			config.bsdauth = 2;
-		else
-			config.bsdauth = 0;
-	}
-#endif
-	if (fparse(config.sslportc, "sslport", config.file,
-	    sizeof(config.sslportc)) == -1)
-		config.ssl = 0;
-	else {
-		config.ssl = 1;
-		config.sslport = atoi(config.sslportc);
-		if (fparse(config.sslcert, "sslcert", config.file,
-		    sizeof(config.sslcert))  == -1) {
-			fprintf(stderr, "Missing <sslcert> in config file.\n");
-			r = -1;
-		}
-		if (fparse(config.sslcertkey, "sslcertkey", config.file,
-		    sizeof(config.sslcertkey)) == -1) {
-			fprintf(stderr, "Missing <sslcertkey> in config file."
-			    "\n");
-			r = -1;
-		}
-	}
-	if (fparse(config.c401, "custom_401", config.file, sizeof(config.user))
-	    == -1)
-		config.c401[0] = '\0';
-	if (fparse(config.c403, "custom_403", config.file, sizeof(config.user))
-	    == -1)
-		config.c403[0] = '\0';
-	if (fparse(config.c404, "custom_404", config.file, sizeof(config.user))
-	    == -1)
-		config.c404[0] = '\0';
-	if (fparse(config.homedirs, "homedirs", config.file,
-	    sizeof(config.homedirs)) == -1)
-		config.homedirs[0] = '\0';
-	if (fparse(config.homedirs_public, "homedirs_public", config.file,
-	    sizeof(config.homedirs_public)) == -1) {
-		config.homedirs_public[0] = '\0';
-	} else {
-		if (config.homedirs_public[0] != '/') {
-			strlcpy(tmp, config.homedirs_public, sizeof(tmp));
-			snprintf(config.homedirs_public,
-			    sizeof(config.homedirs_public), "/%s", tmp);
-		}
-	}
-
-	/* we need all mandatory parameters otherwise no start */
-	if (r == -1)
+	/* parse and load configuration file */
+	if (load_config(0) == -1)
 		exit(1);
 
+	/*
+	 * initialize memory for basic authentication cache
+	 */
+	r = sys_bac_init();
+	if (r == -1)
+		errx(1, "couldn't setup basic authentication cache");
+
 	/*
 	 * change to serverroot directory
 	 */
@@ -672,9 +739,9 @@ main(int argc, char *argv[])
 		snprintf(http_url, sizeof(http_url), "%s:%d",
 		    config.servername, config.serverport);
 	if (config.ssl && config.sslport == PRTS)
-		snprintf(http_urls, sizeof(http_url), "%s", config.servername);
+		snprintf(http_urls, sizeof(http_urls), "%s", config.servername);
 	if (config.ssl && config.sslport != PRTS)
-		snprintf(http_urls, sizeof(http_url), "%s:%d",
+		snprintf(http_urls, sizeof(http_urls), "%s:%d",
 		    config.servername, config.sslport);
 	/* servers connection timeout */
 	snprintf(http_fv_alv, sizeof(http_fv_alv), "timeout=%d, max=%d", TON,
@@ -686,7 +753,7 @@ main(int argc, char *argv[])
 	 * clear all environment variables and save PATH
 	 */
 	strlcpy(http_path, getenv("PATH"), sizeof(http_path));
-	memset(environ, 0, sizeof(environ));
+	*environ = NULL;
 
 	/*
 	 * get servers IP adress(es)
@@ -723,9 +790,9 @@ main(int argc, char *argv[])
 			err(1, "SSL_CTX_new");
 
 		/* load certificate and private key and check them */
-		if (SSL_CTX_use_certificate_file(ssl_ctx, config.sslcert,
-		    SSL_FILETYPE_PEM) != 1)
-			err(1, "SSL_CTX_use_certificate_file");
+		if (SSL_CTX_use_certificate_chain_file(ssl_ctx, config.sslcert)
+		    != 1)
+			err(1, "SSL_CTX_use_certificate_chain_file");
 		if (SSL_CTX_use_PrivateKey_file(ssl_ctx, config.sslcertkey,
 		    SSL_FILETYPE_PEM) != 1)
 			err(1, "SSL_CTX_use_PrivateKey_file");
@@ -1061,15 +1128,9 @@ main(int argc, char *argv[])
 	}
 
 	/*
-	 * signal handler
+	 * setup own signal handler for the parent process
 	 */
-	signal(SIGCHLD, sig_handler);
-	signal(SIGTERM, sig_handler);
-	signal(SIGINT, sig_handler);
-	signal(SIGPIPE, sig_handler);
-	signal(SIGHUP, sig_handler);
-	signal(SIGQUIT, sig_handler);
-	signal(SIGALRM, sig_handler);
+	sys_sighandler(1);
 
 	/*
 	 * open syslog
@@ -1099,9 +1160,14 @@ main(int argc, char *argv[])
 		syslog(LOG_ERR, "open: %s", strerror(errno));
 		exit(1);
 	}
-	dup(0);
-	dup(0);
-
+	if (dup(0) == -1) {
+		syslog(LOG_ERR, "dup: %s", strerror(errno));
+		exit(1);
+	}
+	if (dup(0) == -1) {
+		syslog(LOG_ERR, "dup: %s", strerror(errno));
+		exit(1);
+	}
 #ifdef __OpenBSD__
 	/*
 	 * basic authentication via bsd authentication framework
@@ -1125,7 +1191,6 @@ main(int argc, char *argv[])
 		close(fdbsdauth[1]);
 	}
 #endif
-
 	/*
 	 * chroot to serverroot directory
 	 */
@@ -1405,7 +1470,7 @@ main(int argc, char *argv[])
 				c[sdnow].x_sta = 0;
 				free(h);
 				free(b);
-				continue;
+				goto quit;
 			}
 
 			/* receive data */
@@ -1447,7 +1512,7 @@ main(int argc, char *argv[])
 				c[sdnow].x_sta = 0;
 				free(h);
 				free(b);
-				continue;
+				goto quit;
 			}
 
 			/* post */
@@ -1471,7 +1536,7 @@ main(int argc, char *argv[])
 					c[sdnow].x_sta = 0;
 					free(h);
 					free(b);
-					continue;
+					goto quit;
 				}
 				if (in[i] == '\n' && in[i + 2] == '\n') {
 					tmp[j] = in[i];
