@@ -1,5 +1,7 @@
+/*	$nostromo: proto.h,v 1.46 2016/04/12 19:02:06 hacki Exp $ */
+
 /*
- * Copyright (c) 2004 - 2011 Marcus Glocker <marcus@nazgul.ch>
+ * Copyright (c) 2004 - 2016 Marcus Glocker <marcus@nazgul.ch>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -23,6 +25,7 @@ void		send_full(const int, const int);
 void		send_chunk(const int, const int);
 void		send_file(const int, const int);
 int		sdlisten(int, int);
+int		load_config(int);
 
 /*
  * http.c
@@ -54,6 +57,10 @@ struct header *	http_header(const char *, const char *, const int, const int,
  */
 int		sys_mime(char *, const int, const char *, const int,
 		    const char *);
+int		sys_bac_init(void);
+void		sys_bac_free(void);
+void		sys_bac_add(const char *, const char *);
+int		sys_bac_match(const char *, const char *);
 int		sys_access_auth(const char *, const char *);
 #ifdef __OpenBSD__
 int		sys_access_bsdauth(const char *, const char *);
@@ -71,3 +78,4 @@ void		sys_log(const int, const char *, ...);
 int		sys_compar(const void *, const void *);
 char *		sys_date(struct tm *);
 char *		sys_benv(const char *, ...);
+void		sys_sighandler(int);
