@@ -1,5 +1,7 @@
+/*	$nostromo: sys.c,v 1.74 2022/09/22 15:47:23 hacki Exp $ */
+
 /*
- * Copyright (c) 2004 - 2011 Marcus Glocker <marcus@nazgul.ch>
+ * Copyright (c) 2004 - 2016 Marcus Glocker <marcus@nazgul.ch>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -40,7 +42,7 @@
 #include "../libmy/file.h"
 #ifdef __linux__
 #include <crypt.h>
-#include "../libbsd/strlcpy.h"
+#include "../libmy/strlcpy.h"
 #endif
 #include "proto.h"
 #include "extern.h"
@@ -119,6 +121,128 @@ sys_mime(char *dst, const int dsize, const char *mimes, const int msize,
 	return (0);
 }
 
+/*
+ * sys_bac_init()
+ *	initialize memory for basic authentication cache
+ * Return:
+ *	0 = success, -1 = error
+ */
+int
+sys_bac_init(void)
+{
+	int	i;
+
+	bac = malloc(BAC_ENTRIES * sizeof(*bac));
+	if (bac == NULL)
+		return (-1);
+
+	for (i = 0; i < BAC_ENTRIES; i++) {
+		bac[i] = malloc(sizeof(struct ba_cache));
+		if (bac[i] == NULL) {
+			sys_bac_free();
+			return (-1);
+		}
+		bac[i]->timestamp = 0;
+		memset(bac[i]->username, 0, sizeof(bac[i]->username));
+		memset(bac[i]->password, 0, sizeof(bac[i]->password));
+	}
+
+	return (0);
+}
+
+/*
+ * sys_bac_free()
+ *	free basic authentication cache memory
+ */
+void
+sys_bac_free(void)
+{
+	int 	i;
+
+	for (i = 0; i < BAC_ENTRIES; i++) {
+		if (bac[i] != NULL)
+			free(bac[i]);
+	}
+
+	if (bac != NULL)
+		free(bac);
+}
+
+/*
+ * sys_bac_add()
+ *	add new basic authentication credentials to our cache
+ */
+void
+sys_bac_add(const char *username, const char *password)
+{
+	int	i, i_oldest, already_cached;
+	time_t	t_oldest;
+
+	i = i_oldest = t_oldest = already_cached = 0;
+
+	/* check if user is already cached */
+	for (i = 0; i < BAC_ENTRIES; i++) {
+		if (bac[i]->timestamp == 0)
+			continue;
+		if (strcmp(username, bac[i]->username) == 0) {
+			/* cached entry will be overwritten */
+			already_cached = 1;
+			break;
+		}
+	}
+
+	/* user isn't cached yet */
+	if (already_cached == 0) {
+		/* find empty slot */
+		for (i = 0; i < BAC_ENTRIES; i++) {
+			if (bac[i]->timestamp == 0)
+				break;
+		}
+
+		/* if cache was full rotate by overwriting oldest entry */
+		if (i == BAC_ENTRIES) {
+			time(&t_oldest);
+			for (i = 0; i < BAC_ENTRIES; i++) {
+				if (bac[i]->timestamp < t_oldest) {
+					i_oldest = i;
+					t_oldest = bac[i]->timestamp;
+				}
+			}
+			i = i_oldest;
+		}
+	}
+
+	/* add or update entry */
+	time(&bac[i]->timestamp);
+	strlcpy(bac[i]->username, username, sizeof(bac[i]->username));
+	strlcpy(bac[i]->password, password, sizeof(bac[i]->password));
+}
+
+/*
+ * sys_bac_match()
+ * 	check if basic authentication credentials are cached and if the
+ *	provided password matches with the cached entry
+ * Return:
+ *	0 = credentials don't exist, 1 = credentials exist and password matches
+ */
+int
+sys_bac_match(const char *username, const char *password)
+{
+	int	i;
+
+	for (i = 0; i < BAC_ENTRIES; i++) {
+		if (bac[i]->timestamp == 0)
+			continue;
+		if (strcmp(username, bac[i]->username) == 0 &&
+		    strcmp(password, bac[i]->password) == 0) {
+			/* username and password matches */
+			return (1);
+		}
+	}
+
+	return (0);
+}
+
 /*
  * sys_access_auth()
  *	basic user authentication
@@ -131,10 +255,13 @@ sys_access_auth(const char *clientuser, const char *clientpw)
 	int		i, r, fd, found;
 	char		ch[1], *password;
 	char		serveruserpw[1024], serveruser[1024], serverpw[1024];
-	char		salt[2];
 
 	i = r = found = 0;
 
+	/* check if credentials are already cached */
+	if (sys_bac_match(clientuser, clientpw))
+		return (1);
+
 	memset(serveruserpw, 0, sizeof(serveruserpw));
 
 	/* open htpasswd */
@@ -166,16 +293,19 @@ sys_access_auth(const char *clientuser, const char *clientpw)
 		return (r);
 	}
 
-	/* get DES password salt from htpasswd */
-	salt[0] = serverpw[0];
-	salt[1] = serverpw[1];
+	/* generate encrypted password */
+	password = crypt(clientpw, serverpw);
+	if (password == NULL) {
+		syslog(LOG_ERR, "sys_access_auth: crypt() returned NULL");
+		return (-1);
+	}
 
-	/* generate DES encrypted password */
-	password = crypt(clientpw, salt);
-
-	/* compare DES encrypted passwords */
-	if (strcmp(password, serverpw) == 0)
+	/* compare encrypted passwords */
+	if (strcmp(password, serverpw) == 0) {
+		/* add credentials to our cache */
+		sys_bac_add(clientuser, clientpw);
 		r = 1;
+	}
 
 	/* close htpasswd */
 	close(fd);
@@ -192,24 +322,29 @@ sys_access_auth(const char *clientuser, const char *clientpw)
 int
 sys_access_bsdauth(const char *clientuser, const char *clientpw)
 {
-	int	r, clientuserlen, clientpwlen;
-	char	answer[4];
+	int	clientuserlen, clientpwlen;
+	char	answer[4], msg[1024];
 
-	r = 0;
+	/* check if credentials are already cached */
+	if (sys_bac_match(clientuser, clientpw))
+		return (1);
 
 	clientuserlen = strlen(clientuser);
 	clientpwlen = strlen(clientpw);
 	if (clientuserlen == 0 || clientpwlen == 0)
-		return (r);
+		return (0);
 
-	sys_write(fdbsdauth[0], clientuser, clientuserlen);
-	sys_write(fdbsdauth[0], clientpw, clientpwlen);
+	snprintf(msg, sizeof(msg), "%s %s", clientuser, clientpw);
+	sys_write(fdbsdauth[0], msg, strlen(msg));
 	memset(answer, 0, sizeof(answer));
 	sys_read(fdbsdauth[0], answer, sizeof(answer));
-	if (strcmp(answer, "ok") == 0)
-		r = 1;
+	if (strcmp(answer, "ok") == 0) {
+		/* add credentials to our cache */
+		sys_bac_add(clientuser, clientpw);
+		return (1);
+	}
 
-	return (r);
+	return (0);
 }
 
 /*
@@ -223,27 +358,26 @@ sys_access_bsdauth(const char *clientuser, const char *clientpw)
 void
 sys_daemon_bsdauth(void)
 {
-	int	r;
-	char	clientuser[128], clientpw[128];
+	char	clientuser[128], clientpw[128], msg[1024];
 
-	r = 0;
+	/* nhttpd child processes follow default signal handling */
+	sys_sighandler(0);
 
+	/* main loop */
 	while (!quit) {
-		memset(clientuser, 0, sizeof(clientuser));
-		memset(clientpw, 0, sizeof(clientpw));
-		if (sys_read(fdbsdauth[1], clientuser, sizeof(clientuser)) < 1)
+		if (sys_read(fdbsdauth[1], msg, sizeof(msg)) < 1)
 			continue;
-		if (sys_read(fdbsdauth[1], clientpw, sizeof(clientpw)) < 1)
-			continue;
-		r = auth_userokay(clientuser, NULL, NULL, clientpw);
-		if (r > 0)
+
+		strcutw(clientuser, msg, 1, sizeof(clientuser));
+		strcutw(clientpw, msg, 2, sizeof(clientpw));
+
+		if (auth_userokay(clientuser, NULL, NULL, clientpw))
 			sys_write(fdbsdauth[1], "ok", 2);
 		else
 			sys_write(fdbsdauth[1], "nok", 3);
 	}
 }
 #endif
-
 /*
  * sys_read()
  *	synchronous read
@@ -654,3 +788,31 @@ sys_benv(const char *str, ...)
 
 	return (p);
 }
+
+/*
+ * sys_sighandler()
+ *	setup signal handler for parent or child process
+ */
+void
+sys_sighandler(int parent)
+{
+	if (parent) {
+		/* parent process uses signal handler */
+		signal(SIGCHLD, sig_handler);
+		signal(SIGINT, sig_handler);
+		signal(SIGTERM, sig_handler);
+		signal(SIGHUP, sig_handler);
+		signal(SIGPIPE, SIG_IGN);
+		signal(SIGQUIT, SIG_IGN);
+		signal(SIGALRM, SIG_IGN);
+	} else {
+		/* child process does default behaviour */
+		signal(SIGCHLD, SIG_DFL);
+		signal(SIGINT, SIG_DFL);
+		signal(SIGTERM, SIG_DFL);
+		signal(SIGHUP, SIG_DFL);
+		signal(SIGPIPE, SIG_DFL);
+		signal(SIGQUIT, SIG_DFL);
+		signal(SIGALRM, SIG_DFL);
+	}
+}
